\cast.xpl	30-Jun-2020	Loren Blaney	loren.blaney@gmail.com
\Ray Caster - Use the mouse to explore 3D rooms.
\
\Revisions:
\12-Jan-2006: Original version for the PC.
\ 6-Apr-2014: Modified more than a little for RPi.
\15-Nov-2014: Modified for new GetMouse and GetMouseMove intrinsics.
\ 6-Sep-2015: Fix problem with jerky movement when left mouse button is down.
\25-Feb-2017: Clean up code a little.
\30-Jun-2020: Reverse GetMouseMove Y axis for consistency.

inc	codesr;			\intrinsic code declarations

def	Barrier= 20;		\how close the viewer can get to a wall

int	Cell,			\2-dimensional array containing the floor plan
	MouseDx, MouseDy,	\amount mouse moved since last GetMouseMove call
	Vx, Vy;			\viewer's coordinates in the floor plan
real	Vd;			\viewer's direction (radians: 0=east pi/2=south)

char	SpWall(15, 2+64*64);	\wall panel images (sprites)
char	Image(320*200);		\buffer for copy of video screen

def	Pi = 3.141592654,
	PiOver2 = Pi/2.0,	\90 degrees
	Deg2Rad = Pi/180.0;	\multiplier to convert degrees to radians



proc	Exit(Msg);		\Display (error) message and terminate program
char	Msg;
[SetVid(3);			\restore normal text mode
Text(0, Msg);
CrLf(0);
OpenI(0);			\discard any keystrokes
if GetErr then [];		\discard any pending runtime errors
exit;
];	\Exit



func	Trunc(X);		\Convert positive real to integer by truncation
real	X;
return Fix(X-0.5);

\================================ IMAGE LOADING ================================

proc	LoadWalls;	\Load sprite images from .bmp file and set up color regs
int	Hand,		\file handle
	W, H,		\dimensions of entire image in .bmp file (pixels)
	R, G, B,	\red, green, blue
	X, Y,		\screen coordinates (pixels)
	Y320,		\Y*320 (for efficiency)
	I, T;		\index and temporary scratch
char	Screen(320*200);\temporary memory area for loading images



proc	LoadWall(X0, Y0, W, H, Sp);	\Load a sprite data array from Screen
int	X0, Y0,		\coordinates of sprite in Screen (pixels)
	W, H;		\width and height of sprite (pixels)
char	Sp;		\sprite to load
int	I, X, Y;
[Sp(0):= W;
Sp(1):= H;
I:= 2;
for Y:= Y0 to Y0+H-1 do
    for X:= X0 to X0+W-1 do
	[Sp(I):= Screen(X+Y*320);
	I:= I + 1;
	];
];	\LoadWall



[	\LoadWalls
\Read in a 256-color .bmp file
Trap(false);			\don't trap "file not found" error; we'll do it
Hand:= FOpen("walls.bmp", 0);	\open file for input
if GetErr \#0\ then Exit("walls.bmp not found");
Trap(true);
FSet(Hand, ^I);			\set device 3 to file handle
OpenI(3);

for I:= 0 to 17 do T:= ChIn(3);	\skip unused header info
W:= ChIn(3) + ChIn(3)<<8;	\0..32767 (is ample range)
for I:= 0 to 2-1 do T:= ChIn(3);\skip
H:= ChIn(3) + ChIn(3)<<8;	\0..32767 (is ample range)
for I:= 24 to 53 do T:= ChIn(3);\skip
for I:= 0 to 256-1 do
	[B:= ChIn(3);
	 G:= ChIn(3);
	 R:= ChIn(3);
	 T:= ChIn(3);
	 SetPalette(I, R, G, B);
	];

\Load .bmp image into Screen array
for Y:= H-1 downto 0 do		\.bmp files are upside down!
	[Y320:= 320 * Y;
	for X:= 0 to W-1 do
		Screen(X+Y320):= ChIn(3);
	];
FClose(Hand);			\close handle (so it could be used again)

for I:= 0 to 4 do
	[LoadWall(I*64,  0, 64, 64, @SpWall(I,0));
	 LoadWall(I*64, 64, 64, 64, @SpWall(I+5,0));
	 LoadWall(I*64,128, 64, 64, @SpWall(I+10,0));
	];
];	\LoadWalls

\================================== THE MEAT ===================================

proc	DrawWalls;	\Draw a series of wall panels forming the rooms
int	X, Y,		\coordinates in Image array (pixels)
	Y320X,		\Y*320 + X (for efficiency)
	CBx, CBy,	\cell boundary coordinates
	Wy,		\coordinates in wall (Y direction is vertical)
	Wx1,		\X offset into a wall panel
	WP,		\wall panel code (0..$F)
	Dist1,		\modified distance, used to eliminate fish-eye effect
	Color;		\color of pixel that gets displayed
real	Wx,		\coordinates in wall (Y direction is vertical)
	Vx1, Vy1,	\coordinates of start of ray (hops from cell to cell)
	VA,		\view angle of ray (in radians) relative to viewer (Vd)
	Dv, Dh,		\distances to vertical & horizontal cell boundaries
	Dist,		\distance from viewer to point where ray hits wall
	SinA, CosA;	\sine and cosine of view angle (for efficiency)
int	Face, FaceV, FaceH;
def	N=12, W=8, S=4, E=0;

[for X:= 0 to 320-1 do	\raster fill Image (up & down instead of left & right)
	[\Sweep projected ray from -30 to +30 degrees (sweep from left to right)
	VA:= ATan2(Float(X-160), 277.2);

	\Given viewer's position and ray's direction find Wx and Dist:
	CosA:= Cos(VA+Vd);
	SinA:= Sin(VA+Vd);
	Vx1:= Float(Vx);		\working copy of ray's starting position
	Vy1:= Float(Vy);
	repeat	[			\loop thru cells until ray hits a wall
		if CosA >= 0.0 then			\ray heading toward East
			[CBx:= (Trunc(Vx1)+64) & ~$3F;	\cell's eastern boundary
			FaceV:= W;
			]
		else	[CBx:= (Trunc(Vx1) & ~$3F);	\cell's western boundary
			FaceV:= E;
			];
		Dv:= (Float(CBx)-Vx1)/CosA;		\distance to boundary

		if SinA >= 0.0 then			\ray heading Southward
			[CBy:= (Trunc(Vy1)+64) & ~$3F;	\cell's south boundary
			FaceH:= N;
			]
		else	[CBy:= (Trunc(Vy1) & ~$3F);	\cell's north boundary
			FaceH:= S;
			];
		Dh:= (Float(CBy)-Vy1)/SinA;		\distance to boundary

		if Dv <= Dh then		\use the shorter distance
			[			\toward vertical cell boundary
			Wx:= Vy1 + Dv*SinA;	\intercept on vertical boundary
			if FaceV=E then CBx:= CBx-1;
			Vx1:= Float(CBx);	\continue ray thru next cell
			Vy1:= Wx;
			Face:= FaceV;
			]
		else	[			\toward horizontal cell boundary
			Wx:= Vx1 + Dh*CosA;	\intercept on horz. boundary
			Vx1:= Wx;		\continue ray thru next cell
			if FaceH=S then CBy:= CBy-1;
			Vy1:= Float(CBy);
			Face:= FaceH;
			];
		WP:= Cell(Trunc(Vx1)>>6, Trunc(Vy1)>>6);\wall panel of next cell
		];
	until	WP \#0\;			\ray struck a wall

	Dv:= Vx1 - Float(Vx);			\distance from viewer to wall
	Dh:= Vy1 - Float(Vy);
	Dist:= Sqrt(Dv*Dv + Dh*Dh);
	Dist1:= Fix(Dist * Cos(VA)) /2;		\eliminate fish-eye effect
						\ (use parallel projection)
	Wx:= Wx - 0.5;				\magic crack remover
	Wx1:= Fix(Wx) & $3F;			\horz. position in wall panel
	if Face=E ! Face=N then Wx1:= 63 - Wx1;
	WP:= WP>>Face & $000F;

	Y320X:= X;			\set up column of pixels to be displayed
	for Y:= -100 to 100-1 do
		[Wy:= (Y*Dist1)->>7 + (32-8);	\256 = height scale factor
		case of				\100=200/2; 32=64/2; (centering)
		  Wy<0:  Color:= 7;				\ceiling
		  Wy>63: Color:= $A4				\floor
		other	 Color:= SpWall(WP-1, 2+Wx1+Wy<<6);	\wall panel
		Image(Y320X):= Color;
		Y320X:= Y320X + 320;
		];
	];
];	\DrawWalls

\-------------------------------------------------------------------------------

proc	MoveViewer;
\Move viewer's position (Vx,Vy) and direction (Vd) using the mouse
\ Y axis moves forward and back
\ X axis moves clockwise and counter-clockwise
real	A, Df, Ds, DMax;
int	Mouse,
	Vx6, Vy6,
	CBx, CBy;	\cell boundaries
[Mouse:= GetMouseMove;			\get change in mouse position
MouseDx:= Mouse(0);  MouseDy:= Mouse(1);
Df:= -Float(MouseDy/3);			\Y-axis motion moves along curent dir
Mouse:= GetMouse;
if Mouse(2) & 2 then
	[A:= 0.0;
	Ds:= Float(MouseDx/3);		\X-axis motion moves sideways
	]
else	[\\if Mouse(2) & 1 then [Sound(1, 1, 4542); Sound(1, 1, 4542*2)];
	A:= Float(MouseDx) / 1000.0;	\X-axis motion changes direction
	Ds:= 0.0;
	];
Vd:= Vd + A;
DMax:= Float(Barrier-1) * 0.707;	\speed limit prevents moving thru walls
if Df > DMax then Df:= DMax;
if Df <-DMax then Df:=-DMax;
if Ds > DMax then Ds:= DMax;
if Ds <-DMax then Ds:=-DMax;

\Move viewer and limit how close she can get to a cell containing a block
Vx:= Vx + Fix(Df*Cos(Vd) - Ds*Sin(Vd));	\make tenative East-West move
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6) \#0\ then		\cell to the East contains a block
	if Vx > CBx-Barrier then Vx:= CBx-Barrier;
CBy:= (Vy+64) & ~$3F;
if Cell(Vx6+1, Vy6+1) \#0\ then		\cell to the SE contains a block
	if Vx>CBx-Barrier & Vy>CBy-Barrier then Vx:= CBx-Barrier;
CBy:= Vy & ~$3F;
if Cell(Vx6+1, Vy6-1) \#0\ then		\cell to the NE contains a block
	if Vx>CBx-Barrier & Vy<CBy+Barrier then Vx:= CBx-Barrier;

CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6) \#0\ then		\cell to the West contains a block
	if Vx < CBx+Barrier then Vx:= CBx+Barrier;
CBy:= (Vy+64) & ~$3F;
if Cell(Vx6-1, Vy6+1) \#0\ then		\cell to the SW contains a block
	if Vx<CBx+Barrier & Vy>CBy-Barrier then Vx:= CBx+Barrier;
CBy:= Vy & ~$3F;
if Cell(Vx6-1, Vy6-1) \#0\ then		\cell to the NW contains a block
	if Vx<CBx+Barrier & Vy<CBy+Barrier then Vx:= CBx+Barrier;

Vy:= Vy + Fix(Df*Sin(Vd) + Ds*Cos(Vd));
Vx6:= Vx>>6;
Vy6:= Vy>>6;

CBy:= (Vy+64) & ~$3F;
if Cell(Vx6, Vy6+1) \#0\ then		\cell to the South contains a block
	if Vy > CBy-Barrier then Vy:= CBy-Barrier;
CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6+1) \#0\ then		\cell to the SE contains a block
	if Vy>CBy-Barrier & Vx>CBx-Barrier then Vy:= CBy-Barrier;
CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6+1) \#0\ then		\cell to the SW contains a block
	if Vy>CBy-Barrier & Vx<CBx+Barrier then Vy:= CBy-Barrier;

CBy:= Vy & ~$3F;
if Cell(Vx6, Vy6-1) \#0\ then		\cell to the North contains a block
	if Vy < CBy+Barrier then Vy:= CBy+Barrier;
CBx:= (Vx+64) & ~$3F;
if Cell(Vx6+1, Vy6-1) \#0\ then		\cell to the NE contains a block
	if Vy<CBy+Barrier & Vx>CBx-Barrier then Vy:= CBy+Barrier;
CBx:= Vx & ~$3F;
if Cell(Vx6-1, Vy6-1) \#0\ then		\cell to the NW contains a block
	if Vy<CBy+Barrier & Vx<CBx+Barrier then Vy:= CBy+Barrier;
];	\MoveViewer

\===============================================================================

int	Mouse;
[	\Main
\Array of cells describing the floor plan:
\Each cell is either open (=0) or occupied by a block.
\Each block has 4 walls.
\Each wall is specified by a hex digit 1..F.
\The direction a wall faces is specified by the order of the digits: N,W,S,E.
\For example, the block $1234 has wall image number 2 facing West.
Cell:=				\West                            --------> Y
    [[$0000,$5555,$5555,$555A,$5555,$5555,$5555,$1111,$1112,$1111,$5555], \   |
     [$66B6,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$6555], \   |
     [$66C6,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$6555], \   |
     [$66C6,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$E555], \   |
     [$6666,$1111,    0,$1113,$1111,    0,$1111,    0,    0,    0,$6555], \   v
     [$55D5,    0,    0,    0,$3111,    0,$1111,    0,    0,    0,$6555], \   X
     [$55C5,    0,    0,    0,$3111,    0,    0,    0,    0,    0,$E555],
     [$55C5,    0,    0,    0,$3111,    0,$1111,    0,    0,    0,$6555],
\N\  [$55C5,    0,    0,    0,$4111,    0,$1111,$3434,$3333,$3333,$5555], \South
     [$55C5,    0,    0,    0,    0,    0,$1111,    0,    0,    0,$6555],
     [$55C5,    0,    0,    0,$1111,    0,    0,    0,    0,    0,$6555],
     [$55C5,$3134,$3133,$3133,$1111,$1111,$1111,    0,    0,    0,$6555],
     [$66C6,    0,    0,    0,$3111,    0,    0,    0,    0,    0,$E555],
     [$66C6,    0,$1111,    0,    0,    0,    0,    0,    0,    0,$6555],
     [$66C5,    0,    0,    0,$1111,    0,    0,    0,    0,$5A55,$5555],
     [$66C6,$6565,$65A5,    0,$A666,$5557,$5557,$5557,$5557,$5557,$6666],
     [$66D6,    0,    0,    0,$5677,    0,    0,    0,    0,    0,$6666],
     [$66C6,    0,$A555,$5575,    0,    0,    0,    0,    0,    0,$6666],
     [$66C6,    0,    0,    0,    0,    0,    0,$7777,    0,    0,$E666],
     [$66C6,    0,    0,$A575,    0,    0,    0,$7777,    0,    0,$6666],
     [$66C6,    0,    0,$5558,$5758,$5758,$5778,    0,    0,    0,$6666],
     [$66C6,    0,$A585,    0,    0,    0,    0,$8775,    0,    0,$6666],
     [$66C6,    0,$5585,    0,    0,    0,    0,$8775,    0,    0,$6666],
     [$66C6,    0,$A585,    0,    0,    0,    0,$8775,    0,    0,$E666],
     [$66D6,    0,    0,    0,    0,    0,    0,    0,    0,    0,$6666],
     [$66C6,    0,$A585,    0,    0,    0,    0,$8575,    0,    0,$6666],
     [$66C6,    0,$5555,$585F,$5885,    0,$8855,$5555,$575A,$5755,$6666],
     [$66C6,    0,$A555,    0,$5585,    0,    0,    0,    0,    0,$6666],
     [$66C6,    0,$5555,    0,$5555,$585A,$58A5,    0,    0,    0,$E666],
     [$66C6,    0,$A5A5,    0,    0,    0,    0,    0,    0,    0,$6666],
     [$6696,    0,$5555,    0,    0,    0,    0,    0,    0,    0,$6666],
     [$6666,$5555,$5555,$5555,$AAAA,$5555,$AAAA,$5555,$AAAA,$5555,$6666]];
				\East

if not OpenMouse then Exit("This program requires a mouse.");
SetVid($13);				\320x200 with 256 colors
LoadWalls;

Trap(false);				\ignore div-by-0 error
Vx:= 64+Barrier;   Vy:= 64+Barrier;	\start viewer in north-west corner
Vd:= PiOver2/2.0;			\face south-east

Mouse:= GetMouseMove;			\initialize mouse's starting position
MouseDx:= Mouse(0);  MouseDy:= Mouse(1);

loop	[MoveViewer;
	DrawWalls;			\draw into Image array
	Paint(0, 0, 320, 200, Image, 320);
	if KeyHit then Exit(" ");	\terminate program when a key is hit
	];
];	\Main
